unit xIPMemo;
//******************************************************************************
//       
//******************************************************************************
//        Author : Roman Romanov boss@small-pro.com
//       Project : IP76.RU' 2017 
//      Web-site : small-pro.com
//   Description :       
//                     TMemo   
//                 ,    
//------------------------------------------------------------------------------
// 	     :
// https://msdn.microsoft.com/en-us/library/windows/desktop/bb775464(v=vs.85).aspx
// 	     :
// https://msdn.microsoft.com/de-de/library/windows/desktop/bb761639(v=vs.85).aspx
//------------------------------------------------------------------------------
//   ,           
//******************************************************************************

{$I pro_param.inc}

interface
uses
  Windows, Messages, SysUtils, Classes, Controls, StdCtrls, Graphics
  ;


type
//******************************************************************************
//   
//******************************************************************************
  TxMargin = class (TPersistent)
  protected
    FRect: TRect;
    FOnChange: TNotifyEvent;
    FLockCount: Integer;
    //--------------------------------------------------------------------------    
    procedure DoChange; virtual;
    function GetIntValue (Index : Integer) : Integer; virtual;
    procedure SetIntValue (Index : Integer; Value : Integer); virtual;
    procedure AssignTo (Dest : TPersistent); override;
    //--------------------------------------------------------------------------
    property OnChange : TNotifyEvent read FOnChange write FOnChange;
  public
    Constructor Create; virtual;
    Destructor Destroy; override;
    //--------------------------------------------------------------------------
    procedure SetRect (ARect : TRect);
    //--------------------------------------------------------------------------    
    procedure Lock;
    procedure Unlock;
    function IsLocked : boolean;
  published
    property Top : Integer index 1 read GetIntValue write SetIntValue default 0; 
    property Left : Integer index 2 read GetIntValue write SetIntValue default 0; 
    property Right : Integer index 3 read GetIntValue write SetIntValue default 0; 
    property Bottom : Integer index 4 read GetIntValue write SetIntValue default 0;             
  end;
  
//******************************************************************************
//  TMemo   
//******************************************************************************
  PMemoControl = ^TCustomMemo;
  
  TxIPMemo = class (TMemo)
  protected
    FBitmap : TBitmap;
    FLayout : TTextLayout;
    FMargin : TxMargin;
    FTextHint : string;
    //--------------------------------------------------------------------------
    procedure ObjectChange (Sender : TObject); virtual;
    procedure DoEnter; override;
    procedure DoExit; override;    
    procedure Change; override;
    procedure SetLayout (Value : TTextLayout); virtual;
    procedure SetMargin (Value : TxMargin); virtual; 
    procedure WMSetFont(var Message: TWMSetFont); message WM_SETFONT;
    procedure CMRecreateWnd(var Message: TMessage); message CM_RECREATEWND;    
    procedure PaintWindow(DC: HDC); override;
    procedure SetTextHint (Value : string);
  public
    Constructor Create (AOwner : TComponent); override;
    Destructor Destroy; override;
    function SetRect (ARect : TRect) : boolean; virtual; 
    function SetMemo (AMemo : PMemoControl; AWithFree : boolean = true) : boolean;
    function TrueWordWrap : boolean;
    function CalcRect : TRect; virtual; 
  published
    property Layout : TTextLayout read FLayout write SetLayout default tlTop;
    property Margin : TxMargin read FMargin write SetMargin;
    property TextHint : string read FTextHint write SetTextHint;
  end;

implementation

//******************************************************************************
//   ( xIPDraw)
//******************************************************************************
function CompareRect (r1, r2 : TRect) : boolean;
begin
  result := (r1.Top = r2.Top) and
            (r1.Left = r2.Left) and
            (r1.Right = r2.Right) and
            (r1.Bottom = r2.Bottom);
end;

//******************************************************************************
//     ( xIPDraw)
//******************************************************************************
function WidthRect  (ARect : TRect) : Integer;
begin
  result := ARect.Right - ARect.Left;
end;

function HeightRect (ARect : TRect) : Integer;
begin
  result := ARect.Bottom - ARect.Top;
end;

//******************************************************************************
//    ( xIPDraw)
//******************************************************************************
procedure DoDrawText(DC : HDC; Text:String; var R: TRect; Flags:Word);
begin
  if (Flags and DT_CALCRECT <> 0) and (Text = '')
  then Text := Text + ' ';
  Windows.DrawTextEx(DC, PChar(Text), Length(Text), R, Flags, nil);
end;

//******************************************************************************
//     ( xIPDraw)
//******************************************************************************
function CalcHeight (DC : HDC; AWidth : Integer; AText : string; AWordBreak : boolean) : Integer;
var DS: Integer; 
    tmp : TRect;
begin
  tmp := Rect (0,0,AWidth,100*AWidth);
  DS := DT_EXPANDTABS or DT_LEFT or DT_CALCRECT;
  if AWordBreak then 
    DS := DS or DT_WORDBREAK;
  DoDrawText(DC, AText, tmp, DS);
  result := HeightRect (tmp)+6;
end;

//******************************************************************************
//         ( xIPDraw)
//******************************************************************************
procedure DrawTextEx(Canvas:TCanvas; R:TRect; Text : String;
                     Layout  : TTextLayout = tlCenter; 
                     Alignment : TAlignment = taLeftJustify; 
                     WordWrap : Boolean = false;
                     Clipped : Boolean = false);

const Alignments: array[0..2] of Word = (DT_LEFT, DT_RIGHT,DT_CENTER);
      WordWraps: array[Boolean] of Word = (0, DT_WORDBREAK);

var Rect, CalcRect: TRect;
    DrawStyle: Integer;
    DY:Integer;
begin
  with Canvas do begin
    Brush.Style := bsClear;
    Rect := R;
    DrawStyle := DT_EXPANDTABS or WordWraps[WordWrap] or Alignments[Integer(Alignment)];
    if not Clipped then
      DrawStyle := DrawStyle or DT_NOCLIP;
      
    if Integer(Layout) <> 0 then begin
      CalcRect := Rect;
      DoDrawText(Canvas.Handle,Text,CalcRect, DrawStyle or DT_CALCRECT);
      DY:=Rect.Bottom-CalcRect.Bottom;
      if Integer(Layout)=1 then DY:=DY div 2;
      if DY<0 then DY:=0;
      Rect.Top:=Rect.Top+DY;
    end;
    DoDrawText(Canvas.Handle,Text,Rect,DrawStyle);
  end;
end;


//******************************************************************************
//  
//     TxMargin = class (TPersistent)
//******************************************************************************
Constructor TxMargin.Create; 
begin
  FRect := Rect(0,0,0,0);
  FOnChange := nil;
  FLockCount := 0;
end;

Destructor TxMargin.Destroy; 
begin
  FOnChange := nil;
  inherited Destroy;
end;

procedure TxMargin.DoChange; 
begin
  if IsLocked then exit;
  if assigned (FOnChange) then FOnChange (self);
end;

function TxMargin.GetIntValue (Index : Integer) : Integer;
begin
  case Index of 
    1 : result := FRect.Top;
    2 : result := FRect.Left;
    3 : result := FRect.Right;
    4 : result := FRect.Bottom;
    else result := -1;
  end;
end;

procedure TxMargin.SetIntValue (Index : Integer; Value : Integer);
//------------------------------------------------------------------------------
  function SetVal (var AOld : Integer; ANew : Integer) : boolean;
  begin
    result := AOld <> ANew;
    AOld := ANew;
  end;
//------------------------------------------------------------------------------
var m : boolean;
begin
  case Index of 
    1 : m := SetVal(FRect.Top, Value);
    2 : m := SetVal(FRect.Left, Value);
    3 : m := SetVal(FRect.Right, Value);
    4 : m := SetVal(FRect.Bottom, Value);
    else m := false;
  end;
  if m then 
    DoChange;
end;

procedure TxMargin.SetRect (ARect : TRect);
var m : boolean;
begin
  m := CompareRect (FRect, ARect);
  if not m then exit;
  FRect := ARect;
  DoChange;
end;

procedure TxMargin.Lock;
begin
  FLockCount := FLockCount + 1;
end;

procedure TxMargin.Unlock;
begin
  FLockCount := FLockCount - 1;
end;

function TxMargin.IsLocked : boolean;
begin
  result := FLockCount <> 0;
end;
  
procedure TxMargin.AssignTo (Dest : TPersistent);
begin
  if Dest = nil then inherited AssignTo(Dest)
  else if Dest is TxMargin then begin
    TxMargin(Dest).SetRect(FRect);
    exit;
  end;
  inherited AssignTo(Dest);
end;

//******************************************************************************
//  TMemo   
//     TxIPMemo = class (TMemo)
//******************************************************************************
Constructor TxIPMemo.Create (AOwner : TComponent);
begin
  inherited Create (AOwner);
  ControlState := ControlState + [csCustomPaint];
  FBitmap := TBitmap.Create;
  FMargin := TxMargin.Create;
  FMargin.OnChange := ObjectChange;
  FLayout := tlTop;
end;

Destructor TxIPMemo.Destroy; 
begin
  FreeAndNil (FMargin);
  FreeAndNil (FBitmap);  
  inherited Destroy;
end;

procedure TxIPMemo.ObjectChange (Sender : TObject);
begin
  if Sender = FMargin then 
    Change;
end;

function TxIPMemo.SetRect (ARect : TRect) : boolean;
begin
  result := Perform(EM_SETRECT, 0, LPARAM(@ARect))>0;   
end;

function TxIPMemo.TrueWordWrap : boolean;
begin
  result := not ((Alignment = taLeftJustify) and (ScrollBars in [ssHorizontal, ssBoth])) 
            and (WordWrap or (Alignment <> taLeftJustify));
end;

function TxIPMemo.CalcRect : TRect;
var s : string;
    h : Integer;
    rct : TRect;
begin
  rct := Rect(0,0,ClientWidth,ClientHeight);
  //--       ----------------  
  rct.Top := rct.Top + FMargin.Top;
  rct.Left := rct.Left + FMargin.Left;
  rct.Right := rct.Right - FMargin.Right;
  rct.Bottom := rct.Bottom - FMargin.Bottom;  
  //--     -   ,  ---          
  result := rct;
  if Layout = tlTop then exit;
  //--     ------------------------------
  FBitmap.Canvas.Font.Assign(Font);
  s := Lines.Text;
  //--           
  if (s = '') and (not Focused) then s := TextHint;
  //--    --------------------------------------  
  h := CalcHeight(FBitmap.Canvas.Handle, WidthRect(rct)-2, s,
                  TrueWordWrap);
  //--       -------------
  case FLayout of
    tlCenter : H := rct.Top + (HeightRect(rct) - H) div 2;
    tlBottom : H := rct.Bottom - H;    
  end;
  //--     ------------------------------
  if (H > rct.Top) then 
    rct.Top := H;
  result := rct;
end;

procedure TxIPMemo.DoEnter;
begin
  inherited DoEnter;
  SetRect (CalcRect);
end;

procedure TxIPMemo.DoExit;
begin
  inherited DoExit;
  SetRect (CalcRect);
end;

procedure TxIPMemo.Change; 
begin
  SetRect (CalcRect);  
  inherited Change;
end;

procedure TxIPMemo.SetLayout (Value : TTextLayout);
begin
  if FLayout = Value then exit;
  FLayout := Value;
  Change;
end;

procedure TxIPMemo.SetMargin (Value : TxMargin); 
begin
  FMargin.Assign(Value);
end;

procedure TxIPMemo.WMSetFont(var Message: TWMSetFont); 
begin
  inherited;
  SetRect (CalcRect);
end;
    
procedure TxIPMemo.CMRecreateWnd(var Message: TMessage);
begin
  inherited;
  SetRect (CalcRect);
end;

procedure TxIPMemo.SetTextHint (Value : string);
begin
  if FTextHint = Value then exit;
  FTextHint := Value;
  if not (csLoading in ComponentState) then
    Change;
end;

procedure TxIPMemo.PaintWindow(DC: HDC); 
var rct : TRect; 
    str : string;
    cnv : TCanvas;
begin
  inherited PaintWindow(DC);
  if Focused then exit;
  str := Lines.Text;
  if str <> '' then exit;

  str := FTextHint;
  if str = '' then exit;
  
  rct := CalcRect;
  InflateRect (rct, -2,-2);
  cnv := TCanvas.Create;
  cnv.Handle := DC;
  cnv.Font.Assign(Font);
  cnv.Font.Color := clBtnShadow;
  DrawTextEx(cnv,rct,str,tlTop,Alignment,TrueWordWrap,false);
  cnv.Free;
end;

function TxIPMemo.SetMemo (AMemo : PMemoControl; AWithFree : boolean = true) : boolean;
var s : string;
begin
  result := AMemo <> nil;
  if not result then exit;

  s := Amemo^.Name;
  BoundsRect := AMemo^.BoundsRect;
  Parent := AMemo^.Parent;
  Align := Amemo^.Align;
  Alignment := TxIPMemo(Amemo^).Alignment;

{$IFDEF VER_XE}
  OnMouseEnter := TxIPMemo(Amemo^).OnMouseEnter;
  OnMouseLeave := TxIPMemo(Amemo^).OnMouseLeave;
{$ENDIF}

  BorderStyle := TxIPMemo(Amemo^).BorderStyle;

  Font.Assign(TxIPMemo(Amemo^).Font);
  Color := TxIPMemo(Amemo^).Color;
  Visible := Amemo^.Visible;
  TabOrder := TxIPMemo(Amemo^).TabOrder;
  ScrollBars := TxIPMemo(Amemo^).ScrollBars;
  WantTabs  := TxIPMemo(Amemo^).WantTabs;
  WantReturns := TxIPMemo(Amemo^).WantReturns;
  WordWrap := TxIPMemo(Amemo^).WordWrap;
  
  OnKeyPress := TxIPMemo(Amemo^).OnKeyPress;
  OnKeyDown := TxIPMemo(Amemo^).OnKeyDown;
  OnKeyUp := TxIPMemo(Amemo^).OnKeyUp;
  OnChange := TxIPMemo(Amemo^).OnChange;
  OnClick := TxIPMemo(Amemo^).OnClick;
  OnMouseDown := TxIPMemo(Amemo^).OnMouseDown;
  OnMouseMove := TxIPMemo(Amemo^).OnMouseMove;
  OnMouseUp := TxIPMemo(Amemo^).OnMouseUp;
  OnEnter := TxIPMemo(Amemo^).OnEnter;
  OnExit := TxIPMemo(Amemo^).OnExit;  

  if AWithFree then begin
    Amemo^.Free;
    Name := s;
    AMemo^ := self;
  end; 
end;

end.

